// zombie.cpp : Defines the entry point for the DLL application.
// (C) Copyright 2010 Tony Bruess
// Please see the README

#include "bzfsAPI.h"
#include "plugin_utils.h"
#include "StateDatabase.h"
#include "../src/bzfs/GameKeeper.h"
#include "../src/bzfs/bzfs.h"
#include "../src/bzfs/CmdLineOptions.h"
#include <cstdlib>
#include <iostream>

bool gameInProgress = false;
int timer;
double lastsecond;
std::map<int, bz_eTeamType> teamtracker;

extern CmdLineOptions *clOptions;
extern uint16_t curMaxPlayers;
extern TeamInfo team[NumTeams];

BZ_GET_PLUGIN_VERSION

class zombieHandler : public bz_EventHandler, public bz_CustomSlashCommandHandler
{
public:
	virtual void process(bz_EventData *eventData);
	virtual bool handle(int playerID, bzApiString command, bzApiString message, bzAPIStringList* /*params*/);
};

zombieHandler zombie;

BZF_PLUGIN_CALL int bz_Load(const char* /*commandLine*/)
{
	bz_registerEvent(bz_ePlayerDieEvent, &zombie);
	bz_registerEvent(bz_ePlayerSpawnEvent, &zombie);
	bz_registerEvent(bz_ePlayerJoinEvent, &zombie);
	bz_registerEvent(bz_eGetPlayerSpawnPosEvent, &zombie);
	bz_registerEvent(bz_eFlagDroppedEvent, &zombie);
	bz_registerEvent(bz_eTickEvent, &zombie);
	bz_registerEvent(bz_eGetAutoTeamEvent, &zombie);
	bz_registerEvent(bz_ePlayerPartEvent, &zombie);

	bz_registerCustomSlashCommand ("start", &zombie);
	bz_registerCustomSlashCommand ("stop", &zombie);
	bz_registerCustomSlashCommand ("switch", &zombie);
	bz_registerCustomSlashCommand ("restart", &zombie);
	bz_registerCustomSlashCommand ("join", &zombie);

	bz_debugMessage(4,"zombie plugin loaded");

	return 0;
}

void fixTeamCount()
{
	int playerIndex, teamNum;
	for (teamNum = RogueTeam; teamNum < RabbitTeam; teamNum++)
		team[teamNum].team.size = 0;
	for (playerIndex = 0; playerIndex < curMaxPlayers; playerIndex++)
	{
		GameKeeper::Player *p = GameKeeper::Player::getPlayerByIndex(playerIndex);
		if (p && p->player.isPlaying())
		{
			teamNum = p->player.getTeam();
			if (teamNum == RabbitTeam)
				teamNum = RogueTeam;
			team[teamNum].team.size++;
		}
	}
}

void addPlayer(GameKeeper::Player *playerData, int index)
{
	void *bufStart = getDirectMessageBuffer();
	void *buf = playerData->packPlayerUpdate(bufStart);
	if (playerData->getIndex() == index)
	{
		broadcastMessage(MsgAddPlayer, (char*)buf - (char*)bufStart, bufStart);
	}
	else
	{
		directMessage(index, MsgAddPlayer, (char*)buf - (char*)bufStart, bufStart);
	}
	int teamNum = int(playerData->player.getTeam());
	team[teamNum].team.size++;
	sendTeamUpdate(-1, teamNum);
	fixTeamCount();
}

void removePlayer(int playerIndex)
{
	GameKeeper::Player *playerData = GameKeeper::Player::getPlayerByIndex(playerIndex);
	if (!playerData)
		return;
	void *buf, *bufStart = getDirectMessageBuffer();
	buf = nboPackUByte(bufStart, playerIndex);
	broadcastMessage(MsgRemovePlayer, (char*)buf-(char*)bufStart, bufStart);
	int teamNum = int(playerData->player.getTeam());
	--team[teamNum].team.size;
	sendTeamUpdate(-1, teamNum);
	fixTeamCount();
}

bool switchPlayer(int playerID, bz_eTeamType team)
{
	GameKeeper::Player *playerData = GameKeeper::Player::getPlayerByIndex(playerID);
	if (!playerData)
		return true;
	removePlayer(playerID);
	playerData->player.setTeam((TeamColor)team);
	addPlayer(playerData, playerID);
	return true;
}

bool isAdmin(int playerID)
{
	bz_PlayerRecord* player = bz_getPlayerByIndex(playerID);
	bool admin = player->hasPerm("ban");
	bz_freePlayerRecord(player);
	return admin;
}

int getPlayerByCallsign(bzApiString callsign)
{
	bzAPIIntList* players = bz_newIntList();
	bz_getPlayerIndexList(players);
	int id = -1;
	for(unsigned int i = 0; i < players->size(); i++)
	{
		bz_PlayerRecord* player = bz_getPlayerByIndex(players->get(i));
		bool found = false;
		if(player->callsign == callsign)
		{
			id = (int)i;
			found = true;
		}
		bz_freePlayerRecord(player);
		if(found)
			break;
	}
	bz_deleteIntList(players);
	return id;
}

static void setVar(int playerID, const string& key, const string& value)
{
	void *bufStart = getDirectMessageBuffer();
	void *buf = nboPackUShort(bufStart, 1);
	buf = nboPackUByte (buf, key.length());
	buf = nboPackString(buf, key.c_str(), key.length());
	buf = nboPackUByte (buf, value.length());
	buf = nboPackString(buf, value.c_str(), value.length());
	const int len = (char*)buf - (char*)bufStart;
	directMessage(playerID, MsgSetVar, len, bufStart);
}

void setZombie(int playerID)
{
	setVar(playerID,"_tankSpeed","35");
	setVar(playerID,"_reloadTime","0.0000001");
	setVar(playerID,"_radarLimit","400");
	setVar(playerID,"_tankAngVel","0.785398");
	bz_givePlayerFlag(playerID, "SR", false);
	switchPlayer(playerID, eGreenTeam);
}

void setHuman(int playerID)
{
	setVar(playerID,"_tankSpeed","25");
	setVar(playerID,"_reloadTime","3.5");
	setVar(playerID,"_radarLimit","0");
	setVar(playerID,"_tankAngVel","0.785398");
	bz_removePlayerFlag(playerID);
	switchPlayer(playerID, eRedTeam);
}

void setRogue(int playerID)
{
	setVar(playerID,"_tankSpeed","0.00000001");
	setVar(playerID,"_reloadTime","0.00000001");
	setVar(playerID,"_radarLimit","0");
	setVar(playerID,"_tankAngVel","0.00000001");
	bz_removePlayerFlag(playerID);
	switchPlayer(playerID, eRogueTeam);
}

void setObserver(int playerID)
{
	setVar(playerID,"_tankSpeed","25");
	setVar(playerID,"_reloadTime","3.5");
	setVar(playerID,"_radarLimit","800");
	setVar(playerID,"_tankAngVel","0.785398");
	bz_removePlayerFlag(playerID);
	switchPlayer(playerID, eObservers);
}

void dontMove(int playerID)
{
	setVar(playerID,"_tankSpeed","0.00000001");
	setVar(playerID,"_reloadTime","0.00000001");
	setVar(playerID,"_radarLimit","0");
	setVar(playerID,"_tankAngVel","0.00000001");
	bz_removePlayerFlag(playerID);
}

void zombieHandler::process(bz_EventData *eventData)
{
	switch (eventData->eventType)
	{
	default:
		break;
	case bz_ePlayerDieEvent: //A player dies
	{
		bz_PlayerDieEventData* diedata = (bz_PlayerDieEventData*)eventData;
		// Human killed by zombie
		if(diedata->team == eRedTeam && diedata->killerTeam == eGreenTeam && gameInProgress)
		{
			// Changing them to a zombie
			bz_sendTextMessage(BZ_SERVER, diedata->playerID, "You are now a zombie -- Go kill those humans!");
			teamtracker[diedata->playerID] = eGreenTeam;
		}
		// Zombie killed by human
		if(diedata->team == eGreenTeam && diedata->killerTeam == eRedTeam && gameInProgress)
		{
			// Changing them to a human
			bz_sendTextMessage(BZ_SERVER, diedata->playerID, "You are now a human -- Go kill those zombies!");
			teamtracker[diedata->playerID] = eRedTeam;
		}
	}
	break;
	case bz_ePlayerJoinEvent: //A player joins the game
	{
		bz_PlayerJoinPartEventData* joindata = (bz_PlayerJoinPartEventData*)eventData;
		bz_sendTextMessage(BZ_SERVER, joindata->playerID, "To start the game, wait untill there are at least three players, then type /start");
		if(joindata->team != eObservers)
		{
			dontMove(joindata->playerID);
			teamtracker[joindata->playerID] = eRogueTeam;
		}
		else
		{
			teamtracker[joindata->playerID] = eObservers;
		}

	}
	break;
	case bz_eGetPlayerSpawnPosEvent: //A spawn location is selected
	{
		bz_GetPlayerSpawnPosEventData* spawnposdata = (bz_GetPlayerSpawnPosEventData*)eventData;
		// Zombie
		if(spawnposdata->team == eGreenTeam && gameInProgress) {
			setVar(spawnposdata->playerID,"_reloadTime","0.00001");
			setVar(spawnposdata->playerID,"_tankSpeed","35");
			setVar(spawnposdata->playerID,"_radarLimit","400");
		}
		// Human
		if(spawnposdata->team == eRedTeam && gameInProgress) {
			setVar(spawnposdata->playerID,"_reloadTime","3.5");
			setVar(spawnposdata->playerID,"_tankSpeed","25");
			setVar(spawnposdata->playerID,"_radarLimit","0");
		}
	}
	break;
	case bz_ePlayerSpawnEvent: //A player spawns
	{
		bz_PlayerSpawnEventData* spawndata = (bz_PlayerSpawnEventData*)eventData;

		if(teamtracker[spawndata->playerID] == eGreenTeam)
		{
			setZombie(spawndata->playerID);
		}

		if(teamtracker[spawndata->playerID] == eRedTeam)
		{
			setHuman(spawndata->playerID);
		}

		if(spawndata->team == eGreenTeam && gameInProgress)
		{
			bz_givePlayerFlag(spawndata->playerID,"SR",true);
		}
	}
	break;
	case bz_eFlagDroppedEvent: //A flag is dropped
	{
		bz_FlagDroppedEventData* flagdropdata = (bz_FlagDroppedEventData*)eventData;
		bz_givePlayerFlag(flagdropdata->playerID, flagdropdata->flagType, true);
	}
	break;
	case bz_eGetAutoTeamEvent: //A player's team is selected via autoteam
	{
		bz_GetAutoTeamEventData *autoteamdata = (bz_GetAutoTeamEventData*)eventData;
		if(gameInProgress || autoteamdata->team == eObservers)
		{
			autoteamdata->team = eObservers;
		}
		else
		{
			autoteamdata->team = eRogueTeam;
		}

	}
	break;
	case bz_eTickEvent: //The server ticks
	{
		bz_TickEventData* tickdata = (bz_TickEventData*)eventData;
		if(timer!=0)
		{
			if(tickdata->time-1>lastsecond)
			{
				lastsecond=tickdata->time;
				timer-=1;
				std::string message;
				switch(timer)
				{
				case 0:
					message=std::string("Go!");
					break;
				case 1:
					message=std::string("1");
					break;
				case 2:
					message=std::string("2");
					break;
				case 3:
					message=std::string("3");
					break;
				case 4:
					message=std::string("4");
					break;
				case 5:
					message=std::string("5");
					break;
				case 6:
					message=std::string("6");
					break;
				case 7:
					message=std::string("7");
					break;
				case 8:
					message=std::string("8");
					break;
				case 9:
					message=std::string("9");
					break;
				case 10:
					message=std::string("10");
					break;
				case 11:
					message=std::string("11");
					break;
				case 12:
					message=std::string("12");
					break;
				case 13:
					message=std::string("13 seconds untill game start");
					break;
				case 14:
					message=std::string("Last chance to join! If you are an observer, type /join to join now!");
					break;
				}
				bz_sendTextMessage(BZ_SERVER,BZ_ALLUSERS,message.c_str());
			}
			// Start the game
			if(timer<1)
			{
				timer=0;
				bz_resetBZDBVar("_tankSpeed");
				bz_resetBZDBVar("_reloadTime");
				bz_resetBZDBVar("_tankAngVel");
				bzAPIIntList* playerlist=bz_newIntList();
				bz_getPlayerIndexList(playerlist);
				int num = 0;
				for ( unsigned int i=0; i < playerlist->size(); i++)
				{
					bz_PlayerRecord *updateData;
					updateData = bz_getPlayerByIndex(playerlist->get(i));
					if(updateData->team == eRogueTeam)
					{
						if (num%2 == 0)
						{
							setZombie(updateData->playerID);
						}
						else
						{
							setHuman(updateData->playerID);
						}

						num++;
					}
					bz_killPlayer(updateData->playerID, false, -1);
					bz_freePlayerRecord(updateData);
					//free the player record
				}
				bz_deleteIntList(playerlist);
				gameInProgress = true;
			}
		}
		// Check if teams are in sync with team tracker

		if(gameInProgress)
		{
		bzAPIIntList* playerlist=bz_newIntList();
		bz_getPlayerIndexList(playerlist);
		for ( unsigned int i=0; i < playerlist->size(); i++)
		{
			bz_PlayerRecord *updateData;
			updateData = bz_getPlayerByIndex(playerlist->get(i));

			if(updateData->team != teamtracker[updateData->playerID])
			{
				if(teamtracker[updateData->playerID] == eGreenTeam)
					setZombie(updateData->playerID);
				if(teamtracker[updateData->playerID] == eRedTeam)
					setHuman(updateData->playerID);
			}

			bz_freePlayerRecord(updateData);
			//free the player record
		}

		bz_deleteIntList(playerlist);
		}
		// Check for a winning team
		if((bz_getTeamCount(eRedTeam) == 0 || bz_getTeamCount(eGreenTeam) == 0) && gameInProgress)
		{
			gameInProgress = false;

			bzAPIIntList* playerlist=bz_newIntList();
			bz_getPlayerIndexList(playerlist);
			for ( unsigned int i=0; i < playerlist->size(); i++)
			{
				bz_PlayerRecord *updateData;
				updateData = bz_getPlayerByIndex(playerlist->get(i));

				if(updateData->team != eObservers)
				{
					dontMove(updateData->playerID);
					setRogue(updateData->playerID);
					bz_killPlayer(updateData->playerID, false, 0);
				}

				bz_freePlayerRecord(updateData);
				//free the player record
			}

			bz_deleteIntList(playerlist);

			if(bz_getTeamCount(eRedTeam) == 0)
			{
				bz_sendTextMessage(BZ_SERVER,BZ_ALLUSERS,"The zombies have won! Congragulations!");
			}
			else if (bz_getTeamCount(eGreenTeam) == 0)
			{
				bz_sendTextMessage(BZ_SERVER,BZ_ALLUSERS,"The humans have won! Congragulations!");
			}
		}
	}
	break;

	case bz_ePlayerPartEvent: //A player leaves the game
	{
		bz_PlayerJoinPartEventData* partdata = (bz_PlayerJoinPartEventData*)eventData;

		teamtracker[partdata->playerID] = eNoTeam;
	}
	break;
	}
}

bool zombieHandler::handle(int playerID, bzApiString command, bzApiString message, bzAPIStringList* /*params*/)
{
	if(command == "start")
	{
		bz_PlayerRecord *player = bz_getPlayerByIndex(playerID);
		if(!gameInProgress)
		{
			if(bz_getTeamCount(eRogueTeam) > 2)
			{
				if(timer == 0)
				{
					bz_sendTextMessagef(BZ_SERVER, BZ_ALLUSERS, "%s has started a new game of zombie", player->callsign.c_str());
					timer = 15;
				}
				else
				{
					bz_sendTextMessagef(BZ_SERVER, playerID, "%s, you can not start a countdown when a countdown has been started", player->callsign.c_str());
				}
			}
			else
			{
				bz_sendTextMessagef(BZ_SERVER, playerID, "%s, you can not start a countdown with less then three players", player->callsign.c_str());
			}
		}
		else
		{
		 	bz_sendTextMessagef(BZ_SERVER, playerID, "%s, you can not start a countdown when there is already a game in progress", player->callsign.c_str());
		}
		bz_freePlayerRecord(player);
	}
	else if(command == "stop")
	{
		bz_PlayerRecord *player = bz_getPlayerByIndex(playerID);

		bz_sendTextMessagef(BZ_SERVER, BZ_ALLUSERS, "%s has ended the current game of zombie", player->callsign.c_str());

		bz_freePlayerRecord(player);

		gameInProgress = false;
		timer = 0;

		// Kill all players and make them not move
		bzAPIIntList* playerlist=bz_newIntList();
		bz_getPlayerIndexList(playerlist);

		for ( unsigned int i=0; i < playerlist->size(); i++)
		{
			bz_PlayerRecord *updateData;
			updateData = bz_getPlayerByIndex(playerlist->get(i));

			if(updateData->team != eObservers)
			{
				dontMove(updateData->playerID);
				setRogue(updateData->playerID);
				bz_killPlayer(updateData->playerID, false, 0);
			}

			bz_freePlayerRecord(updateData);
		}
		bz_deleteIntList(playerlist);
	}
	else if(command == "switch")
	{
		bzAPIStringList params;
		params.tokenize(message.c_str(), " ", 0, true);		
		if(isAdmin(playerID))
		{
			if(params.size() >= 2)
			{
				int otherPlayerID = getPlayerByCallsign(params.get(0));
				if(otherPlayerID != -1)
				{
					bz_PlayerRecord *player = bz_getPlayerByIndex(playerID);
					if(params.get(1) == "red" || params.get(1) == "green" || params.get(1) == "rogue" || params.get(1) == "observer")
					{
						if(params.get(1) == "red")
							setHuman(playerID);
						if(params.get(1) == "green")
							setZombie(playerID);
						if(params.get(1) == "rogue")
							setRogue(playerID);
						if(params.get(1) == "observer")
							setObserver(playerID);
					
						bz_sendTextMessagef(BZ_SERVER, eAdministrators, "%s switched %s to the %s team", player->callsign.c_str(), params.get(0).c_str(), params.get(1).c_str());

						bz_freePlayerRecord(player);
					}
					else
					{
						bz_sendTextMessagef(BZ_SERVER, playerID, "Not a valid team: %s", params.get(1).c_str());
					}
				}
				else
				{
					bz_sendTextMessagef(BZ_SERVER, playerID, "No Such Player: %s", params.get(0).c_str());
				}
			}
			else
			{
			 	bz_sendTextMessage(BZ_SERVER, playerID, "Usage: /switch <player> <team>");
			}
		}
		else
		{
			bz_sendTextMessage(BZ_SERVER, playerID, "You do not have permission to switch teams");
		}
		
		return true;
	}
	else if(command == "restart")
	{
	}
	else if(command == "join")
	{
		bz_PlayerRecord *player = bz_getPlayerByIndex(playerID);
		if(player->team == eObservers)
		{
			if(!gameInProgress)
			{
				setRogue(playerID);
				bz_sendTextMessage(BZ_SERVER,playerID,"The game will start shortly. Good luck!");
			}
			else
			{
				bz_sendTextMessage(BZ_SERVER,playerID,"A game is in progress, please wait until it finishes before joining.");
			}
		}
		else
		{
			bz_sendTextMessage(BZ_SERVER,playerID,"You must be an observer to use this command");
		}
	}
	return true;
}

BZF_PLUGIN_CALL int bz_Unload()
{
	bz_removeEvent(bz_ePlayerDieEvent, &zombie);
	bz_removeEvent(bz_ePlayerSpawnEvent, &zombie);
	bz_removeEvent(bz_ePlayerJoinEvent, &zombie);
	bz_removeEvent(bz_eGetPlayerSpawnPosEvent, &zombie);
	bz_removeEvent(bz_eFlagDroppedEvent, &zombie);
	bz_removeEvent(bz_eTickEvent, &zombie);
	bz_removeEvent(bz_eGetAutoTeamEvent, &zombie);
	bz_removeEvent(bz_ePlayerPartEvent, &zombie);

	bz_removeCustomSlashCommand ("start");
	bz_removeCustomSlashCommand ("stop");
	bz_removeCustomSlashCommand ("switch");
	bz_removeCustomSlashCommand ("restart");
	bz_removeCustomSlashCommand ("join");

	bz_debugMessage(4,"zombie plugin unloaded");

	return 0;
}

// Local Variables: ***
// mode:C++ ***
// tab-width: 8 ***
// c-basic-offset: 2 ***
// indent-tabs-mode: t ***
// End: ***
// ex: shiftwidth=2 tabstop=8